/*
Cyber Knight 2: LZ Encoder
*/

#include <stdio.h>

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include <vector>
#include <deque>

using namespace std;

typedef struct
{
	int pos;
	int ptr;
	int len;
} lz_find;


//#define DEBUG_LZ

#define LOOKAHEAD 1

unsigned char buffer[0x10000];
unsigned char file_buffer[0x10000];
int buf_ptr;

///////////////////////////////////////////////////////

unsigned char out_buffer[0x10000];
int out_ptr;

vector<int> table[256];
deque<lz_find> lz_table[4];
int encode_bytes;

int ptr = 0;
int size = 0;

int max_runs = 0;
int max_delta = 0;

int min_match;
int window_size;
int max_match;

int cache_ptr;
int cache_mask;
int run_mask;
int win_mask;
int win_bits;


void Find_LZ( int start, unsigned char byte, int &length, int &pos )
{
	int longest_length = 0;
	int longest_ptr = 0;

	// use cached positions to quickly look through the file
	for( int lcv = 0; lcv < table[ byte ].size(); lcv++ ) {
		int match_length = 0;
		int ptr = table[ byte ][ lcv ];

		// invalid string; stop scanning
		if( ptr >= start )
			break;

		// LZ window restriction
		if( start - ptr > window_size )
			continue;

		// search for longest identical substring
		for( int lcv2 = start; ( lcv2 < start + max_match ) && ( lcv2 < size ); lcv2++ ) {
			// look for a mismatch
			if( buffer[ lcv2 ] != buffer[ ptr ] )
				break;

			// keep looking
			ptr++;
			match_length++;
		}

		// record new long find
		if( longest_length <= match_length ) {
			longest_ptr = table[ byte ][ lcv ];
			longest_length = match_length;
		}
	}

	// output findings
	length = longest_length;
	pos = longest_ptr;
}


void LZ_Encode( int method )
{
	lz_find lz;
	int start;
	int lcv;
	int counter = 0;

	// Step 0: Determine encoding parameters
	min_match = 3;
	switch( method ) {
		case 0:
			max_match = 0x0f + 3;
			window_size = 0xfff;

			run_mask = 0x0f;
			win_mask = 0xf0;
			win_bits = 4;
			cache_ptr = 0xfee;
			cache_mask = 0xfff;
			break;

		case 1:
			max_match = 0x1f + 3;
			window_size = 0x7ff;

			run_mask = 0x1f;
			win_mask = 0xe0;
			win_bits = 3;
			cache_ptr = 0x7de;
			cache_mask = 0x7ff;
			break;

		case 2:
			max_match = 0x3f + 3;
			window_size = 0x3ff;

			run_mask = 0x3f;
			win_mask = 0xc0;
			win_bits = 2;
			cache_ptr = 0x3be;
			cache_mask = 0x3ff;
			break;

		case 3:
			max_match = 0x7f + 3;
			window_size = 0x1ff;

			run_mask = 0x7f;
			win_mask = 0x80;
			win_bits = 1;
			cache_ptr = 0x17e;
			cache_mask = 0x1ff;
			break;
	}

////////////////////////////////////////////////

	// Step 1: Find all LZ matches
	for( lcv = 0; lcv < 256; lcv++ ) {
		table[ lcv ].clear();
	}
	start = 0;
	encode_bytes = 0;
	lz_table[ method ].clear();

	/****************************************/

	// CK2 - init
	memset( &buffer, 0, cache_ptr );
	memcpy( buffer + cache_ptr, file_buffer, size );

	start += cache_ptr;
	size += cache_ptr;
	for( lcv = 0; lcv < cache_ptr; lcv++ ) {
		table[ 0 ].push_back( lcv );
	}

	/****************************************/

	while( start < size ) {
		int future_length[10];
		int future_pos[10];

		int length;
		int pos;

		// Prepare for LZ
		table[ buffer[ start ] ].push_back( start );

		// Go find the longest substring match (and future ones)
		Find_LZ( start, buffer[ start ], length, pos );

		// Slightly increase ratio performance
		for( lcv = 0; lcv < LOOKAHEAD; lcv++ ) {
			start++; table[ buffer[ start ] ].push_back( start );
			Find_LZ( start, buffer[ start ], future_length[ lcv ], future_pos[ lcv ]);
		}

		// Un-do lookahead
		for( lcv = LOOKAHEAD; lcv > 0; lcv-- ) {
			table[ buffer[ start ] ].pop_back(); start--;
			if( future_length[ lcv-1 ] - length >= lcv )
				length = 0;
		}

		if( length >= min_match ) {
			// Found substring match; record and re-do
			lz.pos = start;
			lz.ptr = pos - start;
			lz.len = length;
			
			lz_table[ method ].push_back( lz );

			// Need to add to LZ table
			for( int lcv = 1; lcv < length; lcv++ )
				table[ buffer[ start + lcv ] ].push_back( start + lcv );

			// Fast update
			start += length;
			
			encode_bytes += 2;
		}
		else {
			// No substrings found; try again
			start++;

			encode_bytes++;
		}
	}

	// insert dummy entry
	lz.pos = -1;
	lz_table[ method ].push_back( lz );

	start -= cache_ptr;
	size -= cache_ptr;
}

void LZ_Commit( int method ) {
	int start;
	int lz_ptr;
	int out_byte;
	int out_bits;
	int out_hold;
	lz_find lz;

	// init
	lz_ptr = 0;
	start = 0;
	out_ptr = 0;
	out_byte = 0;
	out_bits = 8;
	out_hold = 0;

	// Step 0: Determine encoding parameters

	switch( method ) {
		case 0:
			max_match = 0x0f + 3;
			window_size = 0xfff;

			run_mask = 0x0f;
			win_mask = 0xf0;
			win_bits = 4;
			cache_ptr = 0xfee;
			cache_mask = 0xfff;
			break;

		case 1:
			max_match = 0x1f + 3;
			window_size = 0x7ff;

			run_mask = 0x1f;
			win_mask = 0xe0;
			win_bits = 3;
			cache_ptr = 0x7de;
			cache_mask = 0x7ff;
			break;

		case 2:
			max_match = 0x3f + 3;
			window_size = 0x3ff;

			run_mask = 0x3f;
			win_mask = 0xc0;
			win_bits = 2;
			cache_ptr = 0x3be;
			cache_mask = 0x3ff;
			break;

		case 3:
			max_match = 0x7f + 3;
			window_size = 0x1ff;

			run_mask = 0x7f;
			win_mask = 0x80;
			win_bits = 1;
			cache_ptr = 0x17e;
			cache_mask = 0x1ff;
			break;
	}

////////////////////////////////////////////////

	start += cache_ptr;
	size += cache_ptr;

	// Step 2: Prepare encoding methods

	// LZ decoder
	out_buffer[ out_ptr++ ] = method;

	// save method byte location
	out_hold = out_ptr++;

	while( start < size ) {
		lz = lz_table[ method ][ lz_ptr ];
		out_byte >>= 1;

		if( lz.pos == start ) {
			// adjust
			lz.ptr += start;
			lz.ptr &= cache_mask;

			// LZ
			out_byte |= ( 0 << 7 );
			out_buffer[ out_ptr++ ] = lz.ptr & 0xff;
			out_buffer[ out_ptr ] = lz.len - 3;
			out_buffer[ out_ptr++ ] |= ( ( lz.ptr >> 8 ) << ( 8-win_bits ) );

			start += lz.len;
			lz_ptr++;
		}
		else {
			// Free byte
			out_byte |= ( 1 << 7 );
			out_buffer[ out_ptr++ ] = buffer[ start ];

			start++;
		}

		// flush out data
		out_bits--;
		if( !out_bits ) {
			out_buffer[ out_hold ] = out_byte;

			out_hold = out_ptr++;
			out_byte = 0;
			out_bits = 8;
		}
	} // end method check

	// flush out data
	while( out_bits-- ) out_byte >>= 1;
	out_buffer[ out_hold ] = out_byte;

	start -= cache_ptr;
	size -= cache_ptr;
}


void LZ_Encode_2( FILE *in, FILE *out, int raw ) {
	int out_size = 0x7fffffff;
	int method = 0;

	// Step 0: Check file size
	fread( file_buffer, 1, 0x10000, in );
	size = ftell( in );

	if( raw == 4 ) {
		// memcpy
		fputc( 0x04, out );
		fwrite( file_buffer, 1, size, out );

		out_ptr = size+1;
		return;
	}

	for( int lcv = 0; lcv < 4; lcv++ ) {
		// Look for smallest encoding size
		LZ_Encode( lcv );

#if 1
		if( encode_bytes < out_size ) { 
			out_size = encode_bytes;
			method = lcv;
		}
#else
		LZ_Commit( lcv );

		if( out_ptr < out_size ) {
			out_size = out_ptr;
			method = lcv;
		}
#endif
	}

	// Final pass
	LZ_Encode( method );
	LZ_Commit( method );
	if( out_ptr > size ) {
		// memcpy
		fputc( 0x04, out );
		fwrite( file_buffer, 1, size, out );

		out_ptr = size+1;
	}
	else
		fwrite( out_buffer, 1, out_ptr, out );
}